Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Make.Supget tbl t returns None if t is not an operator according to table tbl, and it returns the properties of the operator otherwise.