package pratter

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Pratter.SUPPORTSource

Types and utilities on terms that are to be Pratt parsed.

Sourcetype term

The main type of terms, that contains symbols, applications, infix, prefix or postfix operators.

Sourcetype table

The table is used to store available operators.

Sourceval get : table -> term -> (operator * priority) option

get tbl t returns None if t is not an operator according to table tbl, and it returns the properties of the operator otherwise.

Sourceval make_appl : term -> term -> term

make_appl t u returns the application of t to u, sometimes noted @(t, u), or just t u.