Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Pratter.SUPPORTSourceTypes and utilities on terms that are to be Pratt parsed.
The main type of terms, that contains symbols, applications, infix, prefix or postfix operators.
The table is used to store available operators.
get tbl t returns None if t is not an operator according to table tbl, and it returns the properties of the operator otherwise.